#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

/* parse XSB version 0x28 */

int read16_le(unsigned char * buf) {
    return (buf[0])|(buf[1]<<8);
}

long read32_le(unsigned char * buf) {
    return (buf[0])|(buf[1]<<8)|(buf[2]<<16)|(buf[3]<<24);
}

int read16(unsigned char * buf) {
    return (buf[0]<<8)|(buf[1]);
}

long read32(unsigned char * buf) {
    return (buf[0]<<24)|(buf[1]<<16)|(buf[2]<<8)|(buf[3]);
}

void readordie(unsigned char * buf, size_t size, FILE * infile) {
    off_t where = ftell(infile);
    if (fread(buf,1,size,infile)!=size) {fprintf(stderr,"0x%x byte read from 0x%lx failed\n",size,where); exit(-1);}
}

void seekordie(off_t where, FILE * infile) {
    if (fseek(infile,where,SEEK_SET)) {fprintf(stderr,"seek to 0x%lx failed\n",where); exit(-1);}
}

void whine(const char * fmt, ...) {
    va_list ap;
    va_start(ap, fmt);
    vfprintf(stderr,fmt,ap);
    va_end(ap);

    exit(-1);
}

int main(int argc, char ** argv) {
    FILE * infile=NULL;
    unsigned char buf[512];
    int samplecount,entrycount;
    int i;
    off_t banknameoff,segment16,segment3e,segment1a,segment36,segment3a,samplenameoff;
    size_t byteshandled,filesize;

    if (argc!=2) {fprintf(stderr,"no file specified\n"); return -1;}
    infile = fopen(argv[1],"rb");
    if (!infile) {fprintf(stderr,"error opening %s\n",argv[1]); return -1;}

    printf("file: %s\n",argv[1]);
    fseek(infile,0,SEEK_END);
    filesize=ftell(infile);
    fseek(infile,0,SEEK_SET);

    readordie(buf,0x92,infile);

    /* check that it's like what I know */
    if (memcmp(buf,"KBDS\x0\x28",6))
        whine("expected KBDS,0x00,0x28 at 0x0\n");
    if (read16_le(buf+8)!=3)
        whine("expected 0x0003 at 0x8, found 0x%04x\n",read16_le(buf+8));
    if (read32_le(buf+0xc)!=0)
        whine("expected 0 at 0xc, found 0x%08lx\n",read32_le(buf+0xc));
    if (read16_le(buf+0xa)!=read16(buf+0x12))
        whine("inconsistent sample counts, %d!=%d\n",read16_le(buf+0xa),read16(buf+12));
    if (buf[0x11]!=1)
        whine("expected 1 at 0x11, found 0x%02x\n",buf[0x11]);
    if (read16(buf+0x14)!=0)
        whine("expected 0 at 0x14, found 0x%04x\n",read16(buf+0x14));
    if (read32(buf+0x1e)!=-1)
        whine("expected -1 at 0x1e, found 0x%08lx\n",read32(buf+0x1e));
    if (read32(buf+0x26)!=-1)
        whine("expected -1 at 0x26, found 0x%08lx\n",read32(buf+0x26));
    if (read32(buf+0x2a)!=-1)
        whine("expected -1 at 0x2a, found 0x%08lx\n",read32(buf+0x2a));
    if (read32(buf+0x2e)!=-1)
        whine("expected -1 at 0x2e, found 0x%08lx\n",read32(buf+0x2e));
    if (read32(buf+0x32)!=0x92)
        printf("WARN: bank name offset usually at 0x92, found at 0x%08lx\n",read32(buf+0x32));
    if (read32(buf+0x3e)!=0xd2)
        printf("WARN: *0x3e usually 0xd2, found at 0x%08lx\n",read32(buf+0x3e));
    if (read32(buf+0x42)!=0)
        whine("expected 0 at 0x42, found 0x%08lx\n",read32(buf+0x42));
    if (read32(buf+0x46)!=0)
        whine("expected 9 at 0x46, found 0x%08lx\n",read32(buf+0x46));

    printf("header magic: %02x\n",read16_le(buf+6));
    printf("0x8a magic: %08lx\n",read32(buf+0x8a));
    printf("0x8e magic: %08lx\n",read32(buf+0x8e));

    samplecount=read16_le(buf+0xa);
    entrycount=buf[0x10];

    segment16=read32(buf+0x16);
    segment1a=read32(buf+0x1a);
    samplenameoff=read32(buf+0x22);
    banknameoff=read32(buf+0x32);
    segment36=read32(buf+0x36);
    segment3a=read32(buf+0x3a);
    segment3e=read32(buf+0x3e);

    printf("sample count: %d\n",samplecount);
    printf("entry count: %d\n",entrycount);
    printf("bank name (by header): \"%s\"\n",buf+0x4a);

    byteshandled=0x92;

    readordie(buf,0x40,infile);
    printf("bank name: \"%s\"\n",buf);
    byteshandled+=0x40;

    printf("segment3a table:\n");
    for (i=0;i<samplecount;i++) {
        off_t stringoff;
        seekordie(segment3a+i*6,infile);
        readordie(buf,6,infile);
        stringoff=read32(buf);
        printf("\tsample %3d (%d): [0x%08lx] ",i,(short)read16(buf+4),stringoff);
        byteshandled+=6;

        seekordie(stringoff,infile);
        do {
            readordie(buf,1,infile);
            if (buf[0]) printf("%c",buf[0]);
            byteshandled++;
        } while (buf[0]);
        printf("\n");
    }

    printf("segment1a table:\n");
    for (i=0;i<samplecount;i++) {
        off_t entryoff;
        int len;
        int j;
        seekordie(segment1a+i*5,infile);
        readordie(buf,5,infile);
        entryoff=read32(buf+1);
        printf("\t%3d: %02x [0x%08lx] ",i,buf[0],entryoff);
        byteshandled+=5;

        seekordie(entryoff,infile);
        readordie(buf,1,infile);
        byteshandled++;

        if (buf[0]==0) len=0xc;
        else len=0x20;

        for (j=0;j<len;j++) {
            printf("%02x ",buf[0]);
            if (j<len-1) {
                readordie(buf,1,infile);
                byteshandled++;
            }
        }
        printf("\n");
    }

    printf("segment36 table:\n");
    seekordie(segment36,infile);
    for (i=0;i<entrycount;i++) {
        readordie(buf,2,infile);
        printf("\t%3d: %04x\n",i,read16(buf));
        byteshandled+=2;
    }

    printf("file size: %d (0x%x)\n",filesize,filesize);
    printf("bytes handled: %d (0x%x)\n",byteshandled,byteshandled);

    return 0;
}
